/***************************************************************
 * client.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "config.h"

#define ONE_BY_ONE	0

#define USE_RECV_THREAD	0
#define USE_SEND_THREAD	0

#define USE_ADVANCE	0
#if USE_ADVANCE
#define CS_N_ADVANCE	0		/* Non-zero causes abort */
#else
#define CS_N_ADVANCE	0		/* Don't change this */
#endif

#if USE_RECV_THREAD
#include <signal.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#endif
#include <signal.h>
#include <time.h>
#include <math.h>
#include "gggo.h"
#include "global.h"
#include "play.h"
#include "mc.h"
#include "queue.h"
#include "uct.h"
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <string.h>
#include "pack.h"
#include "cs.h"

extern int	gg_count_final_score(ggColor color, int *dame0);
extern uint32_t	uct_get_n_node(void);
extern uint32_t	gg_generation;

/****************************************
 *	Threshold for score to win	*
 ****************************************/
extern inline float	threshold(float x);
extern inline float
threshold(float x) {
#if 0
  return 0.5 * (1 + tanhf(10 * x));
#elif 0
  return (x >= 10) ?  0.5 : (x >= 0) ? 1 : 0;
#else
  return (x > 0) ? 1 : 0;
#endif
}

/************************************************
 *	Client-Server style implementaion of 	*
 *	  Parallel MC/UCT Algorithm		*
 ************************************************/
/************************************************
 *	Client for Network			*
 *	Broadcast a position			*
 ************************************************/

typedef struct {
  int	fd;
  struct sockaddr_in	send_sa, from_sa;
} csSocket;

static const int	yes = 1;
static uint32_t	cs_total_playout;
static csSocket	cs_socket;
static csPacket	cs_packet;

static struct {
  mcPosition	position;
  ggArc		optarc;
  ggUctNode	*leaf;
} cs_advance[CS_N_ADVANCE + 1];

static mcPosition	*cs_position = &cs_advance[0].position;

/* Setup sockets
 */
static void
cs_setup_socket(csSocket *sock, const char *addr, short port) {

if(0) fprintf(stderr, "--> %s(%p,%p,%d)\n", __func__, sock, addr, port);
  if((sock->fd = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0) {
    cs_exit_perror("Client: socket");
  }
  if(setsockopt(sock->fd, SOL_SOCKET, SO_BROADCAST | SO_REUSEADDR,
		  &yes, sizeof yes) < 0) {
    cs_exit_perror("Client: setsockopt");
  }
  memset(&sock->send_sa, 0, sizeof sock->send_sa);
  sock->send_sa.sin_family = AF_INET;
  sock->send_sa.sin_addr.s_addr = inet_addr(addr);
  sock->send_sa.sin_port = htons(port);

  memset(&sock->from_sa, 0, sizeof sock->from_sa);
  sock->from_sa.sin_family = AF_INET;
  sock->from_sa.sin_addr.s_addr = htonl(INADDR_ANY);
#if 1
  if(bind(sock->fd, (struct sockaddr *)&sock->from_sa,
  	sizeof sock->from_sa) < 0) {
    cs_exit_perror("Client: bind");
  }
#endif
if(0) fprintf(stderr, "<-- %s(%p,%p,%d)\n", __func__, sock, addr, port);
  return;
}

static void
cs_receive_log(const char *msg, int l, csPacket *p) {
  mcArc arc = CS_UNPACK(p->optarc);
  uint32_t node = CS_UNPACK(p->node);
  mcMove move = gg_uct_node[node].arc[arc].move;

if(0) fprintf(stderr, "--> %s(%p,%d,%p)\n", __func__, msg, l, p);
  gprintf("Rcvd %s %d/%d bytes ver/type=%d/%d gen/node/po=%d/%d/%d"
	    " move=%1m score=%d"
#if 0
	    " t=%d+%d+%d@%u from %s:%u"
#endif
	    "\n",
	msg, l, CS_UNPACK(p->size), CS_UNPACK(p->version), CS_UNPACK(p->type),
	CS_UNPACK(p->generation), CS_UNPACK(p->node), CS_UNPACK(p->playout),
	move, ((int)CS_UNPACK(p->u.result.score) << 16) >> 16
#if 0
	, cs_timesub(CS_UNPACK(p->t_recv), CS_UNPACK(p->t_send)),
	cs_timesub(CS_UNPACK(p->t_done), CS_UNPACK(p->t_recv)) ,
	cs_timesub(cs_timestamp(), CS_UNPACK(p->t_done)), CS_UNPACK(p->t_send),
	inet_ntoa(cs_from_sa.sin_addr), ntohs(cs_from_sa.sin_port)
#endif
	);
if(0) fprintf(stderr, "<-- %s(%p,%d,%p)\n", __func__, msg, l, p);
 return;
}

#if USE_RECV_THREAD || USE_SEND_THREAD
static int
cs_lock(pthread_mutex_t *lock) {
  int rc = pthread_mutex_lock(lock);
  if(rc > 0)	GG_FATAL(__func__);
  return rc;
}

static int
cs_unlock(pthread_mutex_t *lock) {
  int rc = pthread_mutex_unlock(lock);
  if(rc > 0)	GG_FATAL(__func__);
  return rc;
}
#endif

#if USE_RECV_THREAD

static volatile int	io_nqueue;
static pthread_t	io_thread;
static pthread_mutex_t	io_lock;
static csPacket		io_packet;

static int
cs_recv_update(csThread *self) {
  ssize_t l;
  socklen_t fromlen;
  UNUSED(self);

if(0) gprintf("--> %s()\n", __func__);
  do {
    while(io_nqueue == 0) {
      cs_usleep(1);
    }
    l = 0;
    do {
      fromlen = sizeof cs_socket.from_sa;
      l = recvfrom(cs_socket.fd, &io_packet, sizeof io_packet, 
		    (l == 0) ? 0 : MSG_DONTWAIT,
		    (struct sockaddr *)&cs_socket.from_sa, &fromlen);
      if(l >= 8 && (unsigned)l == CS_UNPACK(io_packet.size) &&
	  CS_UNPACK(io_packet.version) == 1 && CS_UNPACK(io_packet.type) == 2) {
	if(CS_UNPACK(io_packet.generation) == gg_generation &&
	   !uct_is_wrong_node(CS_UNPACK(io_packet.node))) {
if(0) cs_receive_log("", l, &io_packet);
	  break;
	} else {
	  /* Ignore older packets */
if(0) cs_receive_log("OLDER packet", l, &io_packet);
	}
      } else if(l < 0) {
	if(errno == EAGAIN)		break;
	perror("Client: recvfrom");
      } else if(l == 0) {
	/* This means peer is reset which should not occur when broadcast. */
      } else {
      /* Ignore wrong packets */
if(8) cs_receive_log("WRONG packet", l, &io_packet);
      }
    } while(true);
    cs_lock(&io_lock);
    io_nqueue = 1;
    cs_unlock(&io_lock);
  } while(true);
if(0) gprintf("<-- %s()\n", __func__);
  return 0;
}

#else	/* USE_RECV_THREAD */

static int
cs_receive_result(csSocket *sock, uint32_t minnode, bool verbose) {
  int l;
  socklen_t fromlen;

if(0) gprintf("--> %s(%p,%b)\n", __func__, sock, verbose);
  /* Check if received results.
   */
  fromlen = sizeof sock->from_sa;
  l = recvfrom(sock->fd, &cs_packet, sizeof cs_packet,
	  MSG_DONTWAIT, (struct sockaddr *)&sock->from_sa, &fromlen);
  if(l >= 8 && (unsigned)l == CS_UNPACK(cs_packet.size) &&
      CS_UNPACK(cs_packet.version) == 1 && CS_UNPACK(cs_packet.type) == 2) {
    if(CS_UNPACK(cs_packet.generation) == gg_generation &&
	CS_UNPACK(cs_packet.node) > minnode) {
if(0) cs_receive_log("", l, &cs_packet);
    } else {
      /* Ignore older packets */
if(0 && verbose) cs_receive_log("OLDER packet", l, &cs_packet);
      l = -1;
    }
  } else if(l < 0) {
    if(errno == EAGAIN || errno == EINTR) {
      l = -4;
    } else {
      perror("Client: recvfrom");
      l = -3;
    }
  } else if(l == 0) {
    /* This means peer is reset which should not occur when broadcast. */
  } else {
  /* Ignore wrong packets */
if(8 && verbose) cs_receive_log("WRONG packet", l, &cs_packet);
    l = -2;
  }
if(0) gprintf("<-- %s(%p,%b) l=%d\n", __func__, sock, verbose, l);
  return l;
}

#endif	/* USE_RECV_THREAD */

#if USE_SEND_THREAD

static volatile int	io_send_nqueue;
static pthread_t	io_send_thread;
static pthread_mutex_t	io_send_lock;
static mcPosition	io_position;

static void
cs_send_thread(csThread self) {
  int l;
  uint32_t t_send;
  UNUSED(self);

if(0) fprintf(stderr, "--> %s()\n", __func__);
  do {
    while(io_send_nqueue == 0) {
      cs_usleep(1);
    }
    io_position = *cs_position;
    cs_lock(&io_send_lock);
    io_send_nqueue = 0;
    cs_unlock(&io_send_lock);
    l = cs_pack_position(&cs_packet.u.position.position, &io_position)
	    - (uintptr_t)&cs_packet;
if(0) dump_mc_position(&io_position,
	    CS_UNPACK(cs_packet.u.position.first_move), stderr);
    cs_packet.size = CS_PACK((uint16_t)l);
    cs_packet.version = CS_PACK((uint8_t)1);
    cs_packet.type = CS_PACK((uint8_t)1);
    t_send = cs_timestamp();
    cs_packet.t_send = CS_PACK(t_send);
    while(sendto(cs_socket.fd, &cs_packet, l, 0,
      (struct sockaddr *)&cs_socket.send_sa, sizeof cs_socket.send_sa) != l) {
      if(errno != EAGAIN && errno != EINTR) {
	perror("Client: sendto");
      }
      cs_usleep(5);
    }
if(0) gprintf("Sent %d bytes move=%1m gen/node/po=%d/%d/%d at %6uus"
#if 0
      " to %s:%u"
#endif
      "\n",
      CS_UNPACK(cs_packet.size), CS_UNPACK(cs_packet.u.position.first_move),
      CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
      CS_UNPACK(cs_packet.playout), CS_UNPACK(cs_packet.t_send)
#if 0
      , inet_ntoa(cs_socket.send_sa.sin_addr), ntohs(sock->send_sa.sin_port)
#endif
    );
  } while(true);
if(0) fprintf(stderr, "<-- %s()\n", __func__);
}

#else	/* USE_SEND_THREAD */

static void
cs_send_position(mcPosition *p, csSocket *sock,
	  ggUctNode *leaf, ggArc optarc) {
  int l;
  uint32_t t_send;
  mcMove first_move;

if(0) fprintf(stderr, "--> %s(%p,%p,%p,%d)\n", __func__, p, sock, leaf, optarc);
  cs_packet.generation = CS_PACK(gg_generation);
  cs_packet.playout = CS_PACK(cs_total_playout);
  cs_packet.node = CS_PACK((uint32_t)UCT_NTH(leaf));
  cs_packet.optarc = CS_PACK(optarc);
  first_move = leaf->arc[optarc].move;
  cs_packet.u.position.board_size = CS_PACK((uint8_t)gg_board_size);
  cs_packet.u.position.first_move = CS_PACK(first_move);
if(0) dump_mc_position(p, first_move, stderr);
  l = cs_pack_position(&cs_packet.u.position.position, p) -
	  (uintptr_t)&cs_packet;
  cs_packet.size = CS_PACK((uint16_t)l);
  cs_packet.version = CS_PACK((uint8_t)1);
  cs_packet.type = CS_PACK((uint8_t)1);
  t_send = cs_timestamp();
  cs_packet.t_send = CS_PACK(t_send);
  while(sendto(sock->fd, &cs_packet, l, MSG_DONTWAIT,
	  (struct sockaddr *)&sock->send_sa, sizeof sock->send_sa) != l) {
    if(errno != EAGAIN) {
      perror("Client: sendto");
    }
    cs_usleep(5);
  }
if(0) {
  gprintf("Sent %d bytes move=%1m gen/node/po=%d/%d/%d at %6uus",
      CS_UNPACK(cs_packet.size), first_move,
      CS_UNPACK(cs_packet.generation), CS_UNPACK(cs_packet.node),
      CS_UNPACK(cs_packet.playout), CS_UNPACK(cs_packet.t_send));
      if(0) fprintf(stderr, " to %s:%u", inet_ntoa(sock->send_sa.sin_addr),
		      ntohs(sock->send_sa.sin_port));
      putc('\n', stderr);
}
if(0) fprintf(stderr, "<-- %s(%p,%p,%p,%d)\n", __func__, p, sock, leaf, optarc);
  return;
}

#endif	/* USE_SEND_THREAD */

#define gen_rand64	gg_rand64
#include "SFMT.h"
extern void	gg_init_gen_rand(uint32_t seed);

static csThread	cs_thread[GG_MAX_THREAD];
static int	cs_next_ready, cs_next_score;

static void
cs_init_poll(void) {
  cs_next_ready = cs_next_score = 0;
  return;
}

static csThread *
cs_poll_ready(void) {
  csThread *p;
if(0) fprintf(stderr, "--> %s()\n", __func__);
  for(int i = 0; i < gg_n_internal; i += 1) {
    if(cs_next_ready >= gg_n_internal) cs_next_ready = 0;
    p = &cs_thread[cs_next_ready++];
    if(gg_queue_length(&p->ready_queue, NULL) > 0)	goto exit;
  }
  p = NULL;
exit:
if(0) fprintf(stderr, "<-- %s() p=%p\n", __func__, p);
  return p;
}

static csThread *
cs_poll_score(void) {
  csThread *p;
if(0) fprintf(stderr, "--> %s()\n", __func__);
  for(int i = 0; i < gg_n_internal; i += 1) {
    if(cs_next_score >= gg_n_internal) cs_next_score = 0;
    p = &cs_thread[cs_next_score++];
    if(gg_queue_length(&p->score_queue, NULL) > 0)	goto exit;
  }
  p = NULL;
exit:
if(0) fprintf(stderr, "<-- %s() p=%p\n", __func__, p);
  return p;
}

/* Internal Simulation Thread
 */
extern void	mc_init_gen_rand(uint32_t seed);
static void
cs_do_simulation(csThread *self) {
  ggMessage msg;
  int command;
  ggScore score;

if(0) fprintf(stderr, "--> %s(%p)\n", __func__, self);
  assert(self != NULL);
  mc_init_gen_rand(self->seed);

  /* Start main loop */
  do {
    gg_push(cs_message(0, self->id, CS_READY), &self->ready_queue);
    msg = gg_pop(&self->command, NULL);
    command = cs_command(msg);
    switch(command) {
      default:
      	GG_FATAL("Unknown command: Thread[%d] %#x\n", self->id, msg);
      /* Neverreached */
      case CS_HALT:
      break;

      case CS_EXEC:
#if USE_RAVE
	score = mc(&self->position, &self->rave, (mcMove)cs_value(msg));
#else
	score = mc(&self->position, (mcMove)cs_value(msg));
#endif
	gg_push(cs_message(score, self->id, CS_SCORE), &self->score_queue);
	(void)gg_pop(&self->command, NULL);	/* Handshake */
      break;
    }
  } while(command != CS_HALT);
if(0) fprintf(stderr, "<-- %s(%p)\n", __func__, self);
  return;
}

void
cs_init_client(void) {
  int rc;

  /* Setup internal mc simulator thread
   */
  for(int i = 0; i < gg_n_internal; i += 1) {
    cs_thread[i].id = i;
    cs_thread[i].seed = gg_rand64();		/* Don't use rand32() */
    gg_init_queue(&cs_thread[i].ready_queue);
    gg_init_queue(&cs_thread[i].score_queue);
    gg_init_queue(&cs_thread[i].command);
    /* Create client thread */
    rc = pthread_create(&cs_thread[i].thread, NULL,
    	(void *(*)(void *))cs_do_simulation, (void *)&cs_thread[i]);
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create thread[%d] rc=%d -- %s\n", i, rc, __func__);
    }
  }
  if(gg_use_lan) {
    cs_setup_socket(&cs_socket, CS_ADDR, CS_PORT);
#if USE_RECV_THREAD
    io_nqueue = 0;
    rc = pthread_create(&io_thread, NULL,
	(void *(*)(void *))cs_recv_update, (void *)NULL);
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create recv_thread rc=%d -- %s\n", rc, __func__);
    }
    rc = pthread_mutex_init(&io_lock, NULL);
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create io_lock rc=%d -- %s\n", rc, __func__);
    }
#endif
#if USE_SEND_THREAD
    io_send_nqueue = 0;
    rc = pthread_create(&io_send_thread, NULL,
	(void *(*)(void *))cs_send_thread, (void *)NULL);
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create send_thread rc=%d -- %s\n", rc, __func__);
    }
    rc = pthread_mutex_init(&io_send_lock, NULL);
    if(rc != 0) {
      perror(__func__);
      GG_FATAL("Can't create io_send_lock rc=%d -- %s\n", rc, __func__);
    }
#endif
  }
  cs_init_poll();
  return;
}

/* Flush queues of MC servers.
 */
static void
cs_flush_queue(int wait) {
  csThread *thread;

if(0) gprintf("--> %s()\n", __func__);
  cs_usleep(wait);
  while((thread = cs_poll_score()) != NULL) {
    (void)gg_pop(&thread->score_queue, NULL);
    gg_push(cs_message(0, 0, CS_DONE), &thread->command); /* Handshake */
  }
  if(gg_use_lan) {
#if USE_RECV_THREAD
    do {
      if(io_nqueue == 0)		break;
      cs_lock(&io_lock);
      io_nqueue = 0;
      cs_unlock(&io_lock);
    } while(true);
#else
    while(cs_receive_result(&cs_socket, 0, false) != -4) {
      cs_usleep(5);
    }
#endif
  }
if(0) gprintf("<-- %s()\n", __func__);
  return;
}
  
#if USE_ADVANCE

/* Advance buffer for {position, leaf, optarc}.
 */
static volatile int	cs_head = 0, cs_tail = 0;
static volatile bool	cs_is_full = false;

static inline int cs_advance_next(int x)
  { return (x + 1 >= CS_N_ADVANCE + 1) ? 0 : x + 1; }
static inline int
cs_advance_length(void) {
  int l = cs_tail - cs_head;
  if(l < 0)	l += CS_N_ADVANCE + 1;
  assert(l >= 0);
  assert(l < CS_N_ADVANCE + 1);
  return l;
}

static void
cs_push_advance(ggUctNode *leaf, ggArc optarc) {
if(0) gprintf("--> %s(%p,%d) H/T=%d/%d F=%b p=%p\n", __func__,
	leaf, optarc, cs_head, cs_tail, cs_is_full, cs_position);
  assert(!cs_is_full);
  cs_advance[cs_tail].leaf = leaf;
  cs_advance[cs_tail].optarc = optarc;
  cs_tail = cs_advance_next(cs_tail);
  cs_position = &cs_advance[cs_tail].position;
  if(cs_advance_length() >= CS_N_ADVANCE)	cs_is_full = true;
if(0) gprintf("<-- %s(%p,%d) H/T=%d/%d F=%b p=%p\n", __func__,
	leaf, optarc, cs_head, cs_tail, cs_is_full, cs_position);
  return;
}

static ggUctNode *
cs_pop_advance(ggArc *optarc) {
  ggUctNode *leaf;
if(0) gprintf("--> %s(%d) H/T=%d/%d F=%b p=%p\n", __func__,
	*optarc, cs_head, cs_tail, cs_is_full, cs_position);
  cs_head = cs_advance_next(cs_head);
  leaf = cs_advance[cs_head].leaf;
  *optarc = cs_advance[cs_head].optarc;
  cs_position = &cs_advance[cs_head].position;
  cs_is_full = false;
if(0) gprintf("<-- %s(%d) %p H/T=%d/%d F=%b p=%p\n", __func__,
	*optarc, leaf, cs_head, cs_tail, cs_is_full, cs_position);
  return leaf;
}

static void
cs_init_advance(void) {
if(0) gprintf("--> %s()\n", __func__);
  cs_head = 0;
  cs_tail = 0;
  cs_is_full = false;
  cs_position = &cs_advance[0].position;
if(0) gprintf("<-- %s() H/T=%d/%d F=%b p=%p\n", __func__,
		  cs_head, cs_tail, cs_is_full, cs_position);
  return;
}

#endif	/* USE_ADVANCE */

#if USE_EX_DELAY
static struct {
  ggUctNode	*leaf;
  ggArc		optarc;
} ex_delayline[MAX_EX_DELAY];
extern int	gg_n_delay;
#endif

#if 0
static void
cs_copy_rave(mcRave *dst, const mcRave *src) {
  const int n = (int)(dst->n_rave = src->n_rave);
  for(int i = 0; i < n; i += 1) {
    dst->move[i] = src->move[i];
    dst->color[i] = src->color[i];
  }
  return;
}
#endif

/****************
 * 	Timer	*
 ****************/
volatile bool	cs_timeup = false;

uint32_t
cs_do_playout(ggUctNode *root, uint32_t maxplayout) {
  ggUctNode *leaf, *p;
  float v;
  ggScore score;
  ggArc optarc = ~0;
  ggColor color;
  bool in_tree = false;
  csThread *thread = NULL;
  ggMessage msg;
  uint32_t n_external, n_thread, n_tree, rejected;
  uint32_t node, maxdelay = 0, minnode = 0, updating_node;
#if USE_EX_DELAY
  int delay_head = 0, delay_tail = 0;
#endif
#if USE_RAVE
  mcRave mc_rave;
#endif

if(0) gprintf("--> %s(%p,%d,%.2f)\n", __func__, root, maxplayout);

  assert(root != NULL);
  assert(root->position_sp != NULL);
  assert(!root->position_sp->gameend);
  /* Repeat seek optimal move, expand it and evaluate expanded position.
   */
#if USE_RECV_THREAD
  if(gg_use_lan) {
    io_nqueue = 0;
    cs_unlock(&io_lock);
  }
#endif
#if USE_ADVANCE
  cs_init_advance();
#endif
  cs_total_playout = n_external = n_thread = n_tree = 0;
  rejected = 0;

  do {

#if ! ONE_BY_ONE
#if USE_ADVANCE
    leaf = NULL;
    if(cs_is_full)		goto poll_ready;
#else
    if((thread = cs_poll_ready()) == NULL)	goto poll_score;
#endif
#endif
    /* Go down uct tree from root to a leaf node via optimal moves.
     */
if(0) fprintf(stderr, "Client: Start descending tree\n");
    leaf = root;

    do {
      optarc = (ggArc)uct_select_optarc(leaf);
      assert(optarc < leaf->n_arc);
      if((p = leaf->arc[optarc].node) == NULL)	break;
      leaf = p;		/* Go down. */
      assert(leaf != NULL && leaf->position_sp != NULL);
      if(leaf->position_sp->gameend) {
	/* Terminal node, which should be evaluated already. */
	/* Set score, leaf and optarc to the ones of the parent node. */
	/* Set in_tree to skip update_rave(). */
if(0) gprintf("Client: Terminal node, score=%d uct[%d](%C) - uct[%d](%C,%1m)\n",
	  leaf->score, UCT_NTH(leaf), leaf->color, UCT_NTH(leaf->p_node),
	  leaf->p_node->color, leaf->p_node->arc[leaf->p_arc].move);
	score = - leaf->score;
	optarc = leaf->p_arc;
	leaf = leaf->p_node;
	in_tree = true;
	goto update_by_score;
      }
    } while(true);

    /* if visits >= gg_min_visits create a node.
     * Simulate the (position, move) anyway.
     */
    if(leaf->arc[optarc].visits >= gg_min_visits) {
      /* Restore position, play the move, and create a uct node.
       */
      gg_backto_position(leaf->position_sp);
      color = leaf->color;
      gg_do_play_move(leaf->arc[optarc].move, color);
      color = gg_opposit_color(color);
      p = uct_create_node(color, gg_push_position(), leaf, optarc);
      leaf->arc[optarc].node = p;
if(0) gprintf("Client: Create uct[%d](%C) <- uct[%d].arc[%d](%C,%1m)\n",
	    UCT_NTH(p), color, UCT_NTH(leaf), optarc,
	    leaf->color, leaf->arc[optarc].move);

      if(p->position_sp->gameend) {
	/* Newly created node is an end position. */
	/* Keep leaf and optarc. Score is of the leaf. */
	score = gg_count_final_score(leaf->color, NULL);
#if ADAPT_JAPANESE
	score -= ((leaf->color == GG_BLACK) ? 1 : 0);
#endif
	p->score = - score;
#if USE_INDIVIDUAL_FACTOR
	leaf->arc[optarc].factor = 0;		/* Is deteministic */
#endif
if(0) gprintf("Client: Gameend, score=%d uct[%d](%C) - uct[%d](%C,%1m)\n",
	    p->score, UCT_NTH(p), p->color, UCT_NTH(leaf),
	    leaf->color, leaf->arc[optarc].move);
	in_tree = true;
	goto update_by_score;
      }
      leaf = p;
      optarc = (ggArc)uct_select_optarc(leaf);
      assert(optarc < leaf->n_arc);
    }

    /* Set the width of accept window before simulation.
     */
    node = uct_get_n_node();
    assert(node >= 1);
    if(gg_fixed_window != 0) {
      maxdelay = gg_fixed_window;
    } else {
      maxdelay = floor(gg_log_window * logf((float)node));
      if(maxdelay <= 0) maxdelay = 1;
    }
    assert(maxdelay >= 1);
    minnode = (maxdelay >= node) ? 0 : node - maxdelay;

#if USE_EX_DELAY
    /* For experimental use only.
     */
    if(gg_n_delay > 0) {
      ex_delayline[delay_tail].leaf = leaf;
      ex_delayline[delay_tail].optarc = optarc;
      delay_tail = (delay_tail + 1 >= gg_n_delay + 1) ? 0 : delay_tail + 1;
      if(delay_tail != delay_head)	continue;
      leaf = ex_delayline[delay_head].leaf;
      optarc = ex_delayline[delay_head].optarc;
      delay_head = (delay_head + 1 >= gg_n_delay + 1) ? 0 : delay_head + 1;
    }
#endif

    /* Setup the (position, move) or (state, action) pair.
     */
    gg_setup_position(cs_position, leaf->position_sp, leaf->color);
if(0) gprintf("Client: Start simulation uct[%d](%C,%1m)\n",
		  UCT_NTH(leaf), leaf->color, leaf->arc[optarc].move);

    leaf->arc[optarc].ucb *= gg_mod_fpu;

    /* Send the pair to internal simulation threads.
     */
#if USE_ADVANCE
poll_ready:
#endif
    if(maxdelay > 0 && (thread = cs_poll_ready()) != NULL) {
      maxdelay -= 1;
#if USE_ADVANCE
      if(leaf == NULL) {
	leaf = cs_pop_advance(&optarc);
	assert(leaf != NULL);
      }
#endif
      (void)gg_pop(&thread->ready_queue, NULL);
if(0) gprintf("Thread[%d] --> uct[%d](%1m)\n", thread->id,
		      UCT_NTH(leaf), leaf->arc[optarc].move);
      thread->optarc = optarc;
      thread->node = UCT_NTH(leaf);
      thread->position = *cs_position;	/* Copy the position structure */
      gg_push(cs_message(leaf->arc[optarc].move, 0, CS_EXEC), &thread->command);
#if USE_ADVANCE
    } else {
      if(leaf != NULL) {
	cs_push_advance(leaf, optarc);
      }
#endif
    }

    /* Send the pair to external simulation servers.
     */
    if(gg_use_lan) {
#if USE_SEND_THREAD
      if(io_send_nqueue == 0) {
	cs_packet.generation = CS_PACK(gg_generation);
	cs_packet.playout = CS_PACK(cs_total_playout);
	cs_packet.node = CS_PACK((uint32_t)UCT_NTH(leaf));
	cs_packet.optarc = CS_PACK(optarc);
	cs_packet.u.position.board_size = CS_PACK((uint8_t)gg_board_size);
	cs_packet.u.position.first_move = CS_PACK(leaf->arc[optarc].move);
	cs_lock(&io_send_lock);
	io_send_nqueue = 1;
	cs_unlock(&io_send_lock);
      }
#else
      cs_send_position(cs_position, &cs_socket, leaf, optarc);
#endif
    }

    /* Receive results.
     */
poll_score:
    if((thread = cs_poll_score()) != NULL) {
      msg = gg_pop(&thread->score_queue, NULL);
      if(uct_is_wrong_node(thread->node))	continue;
      score = cs_value(msg);
      optarc = thread->optarc;
      updating_node = thread->node;
      leaf = &gg_uct_node[updating_node];
if(0) gprintf("Thread[%d] <-- uct[%d](%1m) score=%d\n", thread->id,
	    UCT_NTH(leaf), leaf->arc[optarc].move, score);
#if USE_RAVE
      mc_rave = thread->rave;	/* Copy whole structure */
#endif
      gg_push(cs_message(0, 0, CS_DONE), &thread->command); /* Handshake */
#if USE_RECV_THREAD
    } else if(gg_use_lan && io_nqueue > 0) {
      cs_packet = io_packet;
      cs_lock(&io_lock);
      io_nqueue = 0;
      cs_unlock(&io_lock);
#else
    } else if(gg_use_lan && cs_receive_result(&cs_socket, minnode, true) > 0) {
#endif
      score = CS_UNPACK(cs_packet.u.result.score);
      optarc = CS_UNPACK(cs_packet.optarc);
      updating_node = CS_UNPACK(cs_packet.node);
      leaf = &gg_uct_node[updating_node];
if(0) gprintf("External: Result score=%d uct[%d](%1m)\n",
	    score, UCT_NTH(leaf), leaf->arc[optarc].move);
#if USE_RAVE
if(0) dump_cs_rave(&cs_packet.u.result.rave, stderr);
      cs_unpack_rave(&mc_rave, &cs_packet.u.result.rave);
if(0) dump_mc_rave(&mc_rave, stderr);
#endif
    } else {
if(0) fprintf(stderr, "No results are outstading\n");
      continue;			/* Goto descend tree */
    }
    in_tree = false;

    /* To ensure UCT converge, too old results are ignored.
     */
    if(updating_node >= minnode && !uct_is_wrong_node(updating_node)) {
      if(score == GG_UNKNOWN_SCORE) {
if(0) fprintf(stderr, "Client: Unknown_score, node=%d%s\n",
	(uint32_t)UCT_NTH(leaf), thread == NULL ? " from LAN" : "");
	v = 0.5;			/* 0.5 means draw. */
      } else {
update_by_score:
	v = threshold((float)score
#if USE_OFFSET
	    - (leaf->color == root->color ? gg_offset : - gg_offset)
#endif
	    + (leaf->color == GG_WHITE ? gg_komi : - gg_komi));
      }
      uct_update_tree(leaf, optarc, v);

if(0) { char temp[20];
gprintf("%-11s Update uct[%d](%C,%1m) v=%.0f(%d)\n",
  (in_tree ? "In Tree:" : thread == NULL ? "External:" :
   (snprintf(&temp[0], sizeof temp, "Thread[%d]:", thread->id), temp)),
  UCT_NTH(leaf), leaf->color, leaf->arc[optarc].move, v, score); }

#if USE_RAVE
      if(! in_tree) {
#if USE_RAVE_UPTO > 0
	if(gg_movenum < USE_RAVE_UPTO) {
#endif
if(0) dump_mc_rave(&mc_rave, stderr);
	  rave_update_tree(leaf, optarc, v, leaf->color, &mc_rave);
#if USE_RAVE_UPTO > 0
	}
#endif
      }
#endif

      /* Update corresponding playout counter.
       */
      if(in_tree) {
      	n_tree += 1;
      } else if(thread != NULL) {
      	n_thread += 1;
      } else {
      	n_external += 1;
      }
      cs_total_playout += 1;

#if USE_INTERNAL_PLAYOUT
      if(n_thread + n_tree >= maxplayout)	break;
#else
      if(n_thread + n_tree + n_external >= maxplayout)	break;
#endif
    } else {
if(0) fprintf(stderr, "Reject update: node=%u minnode=%u\n",
		updating_node, minnode);
      rejected += 1;
    }
    if(gg_eager_update)	goto poll_score;
  } while(!cs_timeup);

  cs_flush_queue(gg_use_lan ? 50000 : 5000);	/* 50ms TODO Depends on size */
if(2) fprintf(stderr, "#playout = %u (%u+%u+%u) reject=%u\n",
       n_thread + n_tree + n_external, n_thread, n_tree, n_external, rejected);
if(0) gprintf("<-- %s\n", __func__);
  return cs_total_playout;
}
