/***************************************************************
 * types.h
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#ifndef _TYPES_H_
#define _TYPES_H_

#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "config.h"

/****************************************
 *	Useful Macros and Functions	*
 ****************************************/
#ifndef MIN
#define MIN(x, y)	((x) <= (y) ? (x) : (y))
#endif
#ifndef MAX
#define MAX(x, y)	((x) >= (y) ? (x) : (y))
#endif
#define CEIL(x, n)	(((x) + ((n) - 1)) & ~((n) - 1))
#define SQR(x)		((x) * (x))
#define UNUSED(x)	((void)(x))
#define TOKEN_TO_STRING(token)	#token
#define NAME2STR(name)	TOKEN_TO_STRING(name)

extern inline float maxf(float x, float y);
extern inline float minf(float x, float y);
extern inline float sqrf(float x);
extern inline int maxi(int x, int y);
extern inline int mini(int x, int y);
extern inline float maxf(float x, float y) { return x >= y ? x : y; }
extern inline float minf(float x, float y) { return x <= y ? x : y; }
extern inline float sqrf(float x) { return x * x; };
extern inline int maxi(int x, int y) { return x >= y ? x : y; }
extern inline int mini(int x, int y) { return x <= y ? x : y; }

/************************
 *	Constants	*
 ************************/
#define GG_N		(GG_MAX_BOARD + 1)
#define GG_N_Y		GG_N
#define GG_N_X		1
#define GG_BOARDMIN	(GG_N + 1)
#define GG_BOARDMAX	(GG_N * GG_N)
#define GG_BOARDSIZE	(GG_N * GG_N + GG_BOARDMIN)
#define GG_MAX_REN	(GG_MAX_BOARD * GG_MAX_BOARD * 4 / 5)
#if SMALLER_BOARD
#define GG_AVE_REN	(GG_MAX_REN / 2)
#else
#define GG_AVE_REN	(GG_MAX_REN / 4)
#endif
#define GG_REN_ROOT	1	/* Zero is used as nil. */
#define GG_MAX_LIBS_REN	5
#define GG_MAX_MOVE	(GG_MAX_BOARD * GG_MAX_BOARD)
#define GG_MAX_MOVENUM	(GG_MAX_MOVE * 2)
#if ! USE_DYNAMIC_ALLOCATE
#if USE_EX_DELAY
#define GG_MAX_UCT_NODE	(GG_MAX_PLAYOUT + GG_MAX_THREAD + 1 + MAX_EX_DELAY)
#else
#define GG_MAX_UCT_NODE	(GG_MAX_PLAYOUT + GG_MAX_THREAD + 1)
#endif
#define GG_MAX_POSITION	(GG_MAX_UCT_NODE + GG_MAX_MOVE)
#endif

enum {
  GG_N_EAST = GG_N_X,
  GG_N_WEST = -GG_N_X,
  GG_N_NORTH = GG_N_Y,
  GG_N_SOUTH = -GG_N_Y,
};

enum {
  GG_EMPTY = 0,
  GG_WHITE,
  GG_BLACK,
  GG_EXTER,
  GG_NUM_COLORS
};
typedef uint8_t		ggColor;

enum {
  GG_PASS_MOVE = 0,
  GG_NO_MOVE,
  GG_RESIGN_MOVE,
};

#define GG_MIN_COLOR	MIN(GG_WHITE, GG_BLACK)
#define GG_MAX_COLOR	MAX(GG_WHITE, GG_BLACK)
#define GG_N_COLOR	(GG_MAX_COLOR - GG_MIN_COLOR + 1)
#define GG_C2X(x)	((x) - GG_MIN_COLOR)

#define GG_POS(x, y)	(GG_N * (y) + (x) + (GG_BOARDMIN))
#define GG_33		GG_POS(2, 2)
#define GG_HOSHI	GG_POS(3, 3)
#define GG_TAKAHOSHI	GG_POS(4, 4)
#define GG_TENGEN	GG_POS(gg_board_size / 2, gg_board_size / 2)

#if GG_BOARDSIZE <= 127
typedef uint8_t		ggUint;
typedef int8_t		ggInt;
#else
typedef uint16_t	ggUint;
typedef int16_t		ggInt;
#endif
typedef	ggUint		ggPoint;
typedef	ggPoint		ggMove;	 /* Move is same as point (coordinate). */
typedef ggUint		ggID;
typedef ggID		ggRenID;
typedef ggUint		ggSize;		/* number of stones */
typedef uint8_t		ggStatus;
typedef uint32_t	ggMark;		/* Not flag but counter for speed */
typedef uint8_t		ggBool;
typedef int8_t		ggDirection;	/* alternative for enum */
typedef	int16_t		ggScore;
#define GG_UNKNOWN_SCORE	((ggScore)0x8000)
typedef	float		ggScoreValue;
typedef ggUint		ggArc;
typedef ggInt		ggSArc;		/* Signed arc */
typedef float		ggTime;
typedef int32_t		ggMessage;
typedef uint32_t	ggRandomSeed;

#if USE_HASH
typedef	uint64_t	ggHash;
#endif

typedef struct {
  ggRenID	next;		/* for free list */
  ggPoint	member;
  ggSize	size;		/* Number of stones */
  ggSize	liberties;
  ggMark	mark;		/* General purpose mark */
  ggColor	color;
  ggPoint	liberty[GG_MAX_LIBS_REN];
} ggRen;			/* TODO Reduce to 8 byte. */

typedef struct {
  ggRenID	id;
  ggPoint	next;
} ggLink;

typedef struct {
  ggLink	ren;
  ggMark	mark;
  ggColor	color;
} ggBoard;

typedef struct {
  ggPoint	point;
  ggColor	color;
} ggKoInfo;

#if PRINT_STATISTICS
extern FILE	*gg_report;
#endif

typedef uint8_t	ggBlockmap;
extern ggBlockmap	gg_block[GG_BOARDSIZE];

#if USE_OFFSET
extern float		gg_offset;
#endif

typedef enum {
  GG_TERRITORY = 1,
  GG_AREA,
} ggScoring;
extern ggScoring	gg_scoring;

typedef enum {
  GG_IS_LEGAL = 0,
  GG_IS_NOT_EMPTY,
  GG_IS_ILLEGAL_KO,
  GG_IS_SUICIDE,
} ggIllegalReason;

/********************************
 *	Position, the world	*
 ********************************/
typedef struct ggPosition {
  struct ggPosition	*previous_sp, *next_sp;
#if USE_HASH
  ggHash	hash;
#endif
  ggMark	mark_num;
  ggMark	mark_ren_num;
  uint16_t	bw_captured[GG_N_COLOR];
  uint16_t	movenum;
  ggKoInfo	board_ko_info;
  ggPoint	previous_move, previous_move2;
  ggRenID	ren_free;
  ggRenID	max_ren_free;
  uint8_t	n_pass;
  ggBool	gameend;
#if USE_EMPTY_LIST
  ggPoint	empty_list0;
#endif
#if USE_EMPTY_LIST
  ggPoint	empty_list[GG_BOARDSIZE];
#endif
  ggBoard	board[GG_BOARDSIZE];
#if USE_ADJACENT_REN
  ggBool	adjacent_ren[GG_MAX_REN][GG_MAX_REN];
#endif
  ggRen		ren[0];
} ggPosition;

typedef struct {
  struct ggPosition	*previous_sp, *next_sp;
#if USE_HASH
  ggHash	hash;
#endif
  ggMark	mark_num;
  ggMark	mark_ren_num;
  uint16_t	bw_captured[GG_N_COLOR];
  uint16_t	movenum;
  ggKoInfo	board_ko_info;
  ggPoint	previous_move, previous_move2;
  ggRenID	ren_free;
  ggRenID	max_ren_free;
  uint8_t	n_pass;
  ggBool	gameend;
#if USE_EMPTY_LIST
  ggPoint	empty_list0;
#endif
#if USE_EMPTY_LIST
  ggPoint	empty_list[GG_BOARDSIZE];
#endif
  ggBoard	board[GG_BOARDSIZE];
#if USE_ADJACENT_REN
  ggBool	adjacent_ren[GG_MAX_REN][GG_MAX_REN];
#endif
  ggRen		ren[GG_AVE_REN];
} ggPositionAve;

/* printutils.h
 */
extern int	mprintf(const char *format, ...);
extern int	gprintf(const char *format, ...);
extern int	gfprintf(FILE *stream, const char *format, ...);
extern void	vgprintf(FILE *stream, const char *format, va_list ap);
#define gputc(c)	putc((c), stderr)
#define GG_FATAL(...) do {			\
  fprintf(stderr, "Abort in %s: ", __func__);	\
  gprintf(__VA_ARGS__);				\
  abort();					\
} while(0)

#endif /* _TYPES_H_ */
