/***************************************************************
 * uct.c
 *
 * Copyright (c) 2007-2008, Hideki Kato & Fixstars Corp. All rights reserved.
 *
 * This program is free software. You may copy, distribute, and modify it
 * under the terms of the License contained in the file COPYING distributed
 * with this package.  This license is the same as the MIT/X Consortium
 * license.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <math.h>
#include <string.h>
#include "gggo.h"
#include "global.h"
#include "queue.h"
#include "mc.h"
#include "uct.h"
#include "cs.h"

#define USE_NEW_RAVE	0
#if USE_NEW_RAVE
#define C_RAVE		0.45
#define C_UCT		0.45
#define RAVE_BIAS	0.05
#endif

/************************
 *	Interface	*
 ************************/
#if USE_DYNAMIC_ALLOCATE
ggUctNode	*gg_uct_node;
static uint32_t	uct_max_node;
#else
ggUctNode	gg_uct_node[GG_MAX_UCT_NODE];
static uint32_t	uct_max_node = GG_MAX_UCT_NODE;
#endif
static uint32_t	uct_n_node = 0;			/* Private */

/********************************
 *	Joseki routines		*
 ********************************/
static const int gg_trans_matrix[8][4] = {
  {  1,  0,  0,  1 },
  {  0, -1,  1,  0 },
  {  0,  1, -1,  0 },
  { -1,  0,  0, -1 },

  {  1,  0,  0, -1 },
  {  0,  1,  1,  0 },
  {  0, -1, -1,  0 },
  { -1,  0,  0,  1 },
};

static ggMove
gg_symmetry(const unsigned int n, int x, int y) {
  ggMove m;
  int u, v;
  const int center = (gg_board_size - 1) / 2;

if(0) gprintf("--> %s(%1m(%d,%d))\n", __func__, GG_POS(x, y), x, y);
  x -= center;
  y -= center;
  u = gg_trans_matrix[n][0] * x + gg_trans_matrix[n][1] * y + center;
  v = gg_trans_matrix[n][2] * x + gg_trans_matrix[n][3] * y + center;
  m = GG_POS(u, v);
if(0) gprintf("<-- %s=%1m(%d,%d)\n", __func__, m, u, v);
  return m;
}

ggMove
gg_genjoseki(int movenum) {
  ggMove m, p;

if(0) gprintf("--> %s(%d)\n", __func__, movenum);
  m = GG_NO_MOVE;
  /* For 9 x 9 or smaller. */
  if(gg_board_size == 9) {
    if(movenum == 0) {
      m = GG_TENGEN;
#if 0
    } else if(movenum == 1) {
      if(gg_get_color(GG_TENGEN) == GG_BLACK) {
      	m = GG_POS(4, 2);
      }
    } else if(movenum == 2) {
      if(gg_get_color(GG_TENGEN) == GG_BLACK) {
	for(int i = 0; i < 4; i += 1) {
	  p = gg_symmetry(i, 4, 2);
if(0) gprintf(" i=%d p=%1m.\n", i, p);
	  if(gg_get_color(p) == GG_WHITE) {
	    m = gg_symmetry(i, 4, 6);
	    break;
	  }
	}
      }
#endif
    }
    goto exit;
  }
  /* Look for an empty corner. */
  for(int i = 0; i < 4; i += 1) {
    m = gg_symmetry(i, 3, 3);
    if(! gg_is_empty(m))	goto next_corner;
    GG_FOR_8_ADJACENT_POINT(p, m) {
      if(! gg_is_empty(p))	goto next_corner;
    }
    /* Found */
#if SMALLER_BOARD
    m = gg_symmetry(i, 2, 2);
#endif
    goto found;
next_corner:
    continue;
  }
  /* All corners are occupied. */
  m = GG_NO_MOVE;
  goto exit;
found:
  goto exit;
exit:
if(0) gprintf("<-- %s(%d)=%1m\n", __func__, movenum, m);
  return m;
}

/************************************************
 *	Block mode by common fate graph		*
 ************************************************/
ggBlockmap	gg_block[GG_BOARDSIZE];
static ggBlockmap	gg_full_blockmap[GG_BOARDSIZE];

void	gg_init_block(void);
void
gg_init_block(void) {
  memset(&gg_full_blockmap[0], 1, sizeof gg_full_blockmap);
  return;
}

static inline void
gg_clear_block(void) {
  memset(&gg_block[0], 0, sizeof gg_block);
  return;
}

static inline void
gg_mark_block(ggMove p) {
  gg_block[p] = 1;
  return;
}

static inline bool
gg_is_marked_block(ggMove p) {
  assert(gg_is_not_ob(p));
  return gg_block[p] != 0;
}

static inline bool
gg_is_marked_block2(ggMove p, ggBlockmap *map) {
  assert(gg_is_not_ob(p));
  return map[p] != 0;
}

static void
gg_mark_cf(ggMove p, int d) {
  ggMove q;
  int penalty;

  if(gg_block[p] < d) {
    gg_block[p] = d;
    penalty = 0;
    switch(gg_get_color(p)) {
      case GG_EMPTY:
	penalty = 1;
	if(d - penalty <= 0)	break;
	/* Fall through */
      case GG_BLACK:
      case GG_WHITE:
	GG_FOR_8_ADJACENT_POINT(q, p) {
	  gg_mark_cf(q, d - penalty);
	}
      case GG_EXTER:
      break;
      default:
	GG_FATAL("Impossible color at %1m, %d\n", p, gg_get_color(p));
      break;
    }
  }
  return;
}

#if (BLOCK_MODE & 0x1) != 0
#if ADD_BLOCK
static void
gg_mark_around(ggMove m) {
  ggMove p;

  if(gg_is_empty(m)) {
    GG_FOR_8_ADJACENT_POINT(p, m) {
      if(!gg_is_empty(p))	goto exit;
    }
    gg_mark_block(m);
    GG_FOR_8_ADJACENT_POINT(p, m) {
      gg_mark_block(p);
    }
  }
exit:
  return;
}
#endif
#endif

/*
 * BLOCK_MODE
 *  0: No blockmode at all
 *  1: Cfg
 *  2: Openning
 *  3: Cfg with openning (standard)
 */
static ggBlockmap *
gg_setup_block(ggMove lastmove1, ggMove lastmove2) {
  ggBlockmap *map;
  ggMove m;

if(0) gprintf("--> %s(%1m,%1m)\n", __func__, lastmove1, lastmove2);

if(0) {
  gprintf("gg_board:");
  for(m = 0; m < GG_BOARDSIZE; m += 1) {
    gprintf(" %1d", gg_is_b_or_w(m));
  }
  gprintf("\n");
}

#if BLOCK_EXCLUDE_KO
  if(gg_board_ko_info.color != GG_EMPTY) {
    map = &gg_full_blockmap[0];
    goto exit;
  }
#endif

#if BLOCK_MODE == 0
  map = &gg_full_blockmap[0];
  goto exit;
#endif

#if (BLOCK_MODE & 0x2) != 0
  if(lastmove2 == GG_NO_MOVE &&
      (lastmove1 == GG_NO_MOVE || lastmove1 == GG_TENGEN)) {
    /* Mark from 3-3 to TENGEN (1/8). */
    map = &gg_block[0];
    gg_clear_block();
    for(int y = 2; y <= gg_board_size / 2; y += 1) {
      for(int x = y; x <= gg_board_size / 2; x += 1) {
      	gg_mark_block(GG_POS(x, y));
      }
    }
  } else {
    map = &gg_full_blockmap[0];
  }
#endif
#if BLOCK_MODE == 2
  goto exit;
#endif

#if (BLOCK_MODE & 0x1) != 0
#if BLOCK_MODE == 1
  map = &gg_block[0];
  gg_clear_block();
#endif
  if(gg_is_onboard(lastmove1)) {
    gg_mark_cf(lastmove1, BLOCK_DISTANCE + 1);
  }
  if(gg_is_onboard(lastmove2)) {
    gg_mark_cf(lastmove2, BLOCK_DISTANCE + 1);
  }
#if ADD_BLOCK
  /* Mark all empty corners. */
  for(int i = 0; i < 4; i += 1) {
    if(gg_board_size <= 11) {
      m = gg_symmetry(i, 2, 2);
    } else {
      m = gg_symmetry(i, 3, 3);
    }
    gg_mark_around(m);
  }
#if 0
  gg_mark_around(GG_TENGEN);
#endif
#endif
#endif

exit:
if(0) {
  gprintf("gg_block:");
  for(int i = 0; i < GG_BOARDSIZE; i += 1) {
    gprintf(" %1d", gg_block[i]);
  }
  gprintf("\n");
}
if(0) gprintf("<-- %s(%1m,%1m)\n", __func__, lastmove1, lastmove2);
  return map;
}

#if 0
static bool
gg_is_eye_filling(ggMove m, ggColor c) {
  bool	rc;
  union {
    uint8_t	b[4];
    uint32_t	i;
  } neighbor, diagonal;

  assert(!gg_is_illegal(m, c));
#if 0
  for(int i = 0; i < 4; i += 1) {
    neighbor.b[i] = 0;
    diagonal.b[i] = 0;
  }
#else
  neighbor.i = 0;
  diagonal.i = 0;
#endif
  neighbor.b[gg_get_color(GG_NORTH(m))]	+= 1;
  neighbor.b[gg_get_color(GG_EAST(m))]	+= 1;
  neighbor.b[gg_get_color(GG_SOUTH(m))]	+= 1;
  neighbor.b[gg_get_color(GG_WEST(m))]	+= 1;
  diagonal.b[gg_get_color(GG_NE(m))]	+= 1;
  diagonal.b[gg_get_color(GG_SE(m))]	+= 1;
  diagonal.b[gg_get_color(GG_SW(m))]	+= 1;
  diagonal.b[gg_get_color(GG_NW(m))]	+= 1;

#if 1
  rc = neighbor.b[c] + neighbor.b[GG_EXTER] >= 4 &&
   (diagonal.b[gg_opposit_color(c)] + (diagonal.b[GG_EXTER] > 0 ? 1 : 0)) <= 2;
#else
{ ggMove p; ggRenID s;
  rc = false;
  if(neighbor.b[c] + neighbor.b[GG_EXTER] < 4)	goto exit;
#if 0
  if(diagonal.b[gg_opposit_color(c)] > 1)	goto exit;
#else
  if(diagonal.b[gg_opposit_color(c)] > 0)	goto exit;
#endif
  GG_FOR_4_ADJACENT_POINT(p, m) {
    if(!gg_is_friend(p, c))	continue;
    s = gg_id_ren(p);
    assert(s != 0);
    if(gg_libs_ren(s) <= 1)	goto exit;
  }
  rc = true;
}
exit:
#endif
  return rc;
}
#else
#define GG_FOR_4_DIAGONAL_POINT(q, p) \
  for(int16_t _z = 1; (q) = (p) + gg_direction[_z], _z < 8; _z += 2)

static bool
gg_is_eye_filling(ggMove m, ggColor c) {
  bool	rc;
  ggMove p;
  int n_opp;
  bool is_exter;
  ggRenID s;

  assert(!gg_is_illegal(m, c));
  rc = false;
  GG_FOR_4_ADJACENT_POINT(p, m) {
    if(gg_is_friend(p, c)) {
      s = gg_id_ren(p);
      assert(s != 0);
      if(gg_libs_ren(s) <= 1)	goto exit;
    } else if(!gg_is_exter(p)) {
      goto exit;
    }
  }
  n_opp = 0;
  is_exter = false;
  GG_FOR_4_DIAGONAL_POINT(p, m) {
    if(gg_is_exter(p))	{
      is_exter = true;
    } else if(gg_is_opponent(p, c)) {
      s = gg_id_ren(p);
      assert(s != 0);
      if(gg_libs_ren(s) <= 1) {
      	n_opp += 1;
      } else {
	n_opp += 2;
      }
    }
  }
  if(n_opp + (is_exter ? 2 : 0) >= 4)	goto exit;
  rc = true;
exit:
  return rc;
}
#endif

#if PRIOR_LIBERTIES
static inline int
gg_count_libs(ggMove p) {
  return (gg_is_empty(GG_NORTH(p)) ? 1 : 0)
       + (gg_is_empty(GG_EAST(p)) ? 1 : 0)
       + (gg_is_empty(GG_SOUTH(p)) ? 1 : 0)
       + (gg_is_empty(GG_WEST(p)) ? 1 : 0);
}
#endif

/********************************
 *		UCT		*
 ********************************/
#if USE_DYNAMIC_ALLOCATE
extern char *	gg_init_uct(uint32_t n);
char *
gg_init_uct(uint32_t n) {
  int rc;

  uct_max_node = n;
  rc = posix_memalign((void **)&gg_uct_node, sysconf(_SC_PAGESIZE),
			  (size_t)n * sizeof(ggUctNode));
if(2) gprintf("Allocate uct stack: size=%ldMiB top=%p limit=%p\n",
	(n * sizeof(ggUctNode) + ((1L << 20) - 1)) >> 20,
	&gg_uct_node[0], &gg_uct_node[n]);
  return rc == 0 ? NULL : strerror(rc);
}
#endif

uint32_t
uct_get_n_node(void) {
  return uct_n_node;
}

void
uct_init_tree(void) {
  uct_n_node = 0;
  return;
}

bool
uct_is_wrong_node(uint32_t node) {
  return uct_n_node <= node;
}

static inline void
uct_init_arc(ggUctNode *p, int arc, ggMove move, float q, float m) {
  p->arc[arc].node = NULL;
  p->arc[arc].move = move;
  p->arc[arc].visits = m;
  p->arc[arc].value = q;
#if USE_UCB1_TUNED
  p->arc[arc].value2 = SQR(q);
#endif
#if USE_RAVE
  p->arc[arc].m_rave = m;
  p->arc[arc].q_rave = q;
#if USE_UCB1_TUNED
  p->arc[arc].q_rave_sqr = SQR(q);
#endif
#endif
#if USE_INDIVIDUAL_FACTOR
  p->arc[arc].factor = gg_ucb_factor;
#endif
  return;
}

static inline float
uct_ucb(ggUctNode *p, int arc) {
#if NO_VARIANT
  return p->arc[arc].value;
#else
  float log_n_div_n, q, var;

  assert(p->arc[arc].visits > 0);
  assert(p->visits > 0);
  log_n_div_n = logf(p->visits) / p->arc[arc].visits;
  q = p->arc[arc].value;
#if USE_UCB1_TUNED
{
  float v = p->arc[arc].value2 - SQR(q) + sqrtf(2 * log_n_div_n);
  var = sqrtf(minf(0.25, v) * log_n_div_n);
}
#else
#if USE_INDIVIDUAL_FACTOR
  var = p->arc[arc].factor * sqrtf(log_n_div_n);
#else
  var = gg_ucb_factor * sqrtf(log_n_div_n);
#endif
#endif
  return q + var;
#endif
}

#if USE_RAVE
static inline float
uct_ucb_rave(ggUctNode *p, int arc) {
#if NO_VARIANT
  return p->arc[arc].q_rave;
#else
  float log_n_div_n, q, var;

  assert(p->m_rave > 0 && p->arc[arc].m_rave > 0);
  log_n_div_n = logf(p->m_rave) / p->arc[arc].m_rave;
  q = p->arc[arc].q_rave;
#if USE_UCB1_TUNED
{
  float v = p->arc[arc].q_rave_sqr - SQR(q) + sqrtf(2 * log_n_div_n);
  var = sqrtf(minf(0.25, v) * log_n_div_n);
}
#else
#if USE_INDIVIDUAL_FACTOR
  var = p->arc[arc].factor * sqrtf(log_n_div_n);
#else
  var = gg_ucb_factor * sqrtf(log_n_div_n);
#endif
#endif
  return q + var;
#endif
}
#endif

ggUctNode *
uct_create_node(ggColor color, ggPosition *sp, ggUctNode *p_node, ggArc p_arc) {
  ggUctNode *p;
  ggBlockmap *blockmap;
  uint32_t node;
  int n;
  ggMove move;

if(0) gprintf("--> %s(%C,%p,%p,%u)\n", __func__, color, sp, p_node, p_arc);
  node = uct_n_node++;		/* ATOMIC */
  if(node >= uct_max_node)	GG_FATAL("Uct_node[] overflow! %d\n", node);
  p = &gg_uct_node[node];
  p->color = color;
  p->position_sp = sp;
  p->visits = 0;
  p->p_node = p_node;
  p->p_arc = p_arc;
#if USE_RAVE
  p->m_rave = 0;
#endif
#if SAFETY
  p->score = 0;
#endif
#if USE_ARC_CACHE
  p->optarc = -1;
#endif

  /* Generate and store legal moves from center.
   */
if(0) {
  gprintf("Empty point: ");
  GG_FOR_ALL_EMPTY_POINT(move) gprintf(" %1m", move);
  gprintf("\n");
}
  n = 0;
  blockmap = gg_setup_block(gg_previous_move, gg_previous_move2);
  GG_FOR_ALL_EMPTY_POINT(move) {
    if(!gg_is_marked_block2(move, blockmap))	continue;
    if(gg_is_illegal(move, color) || gg_is_eye_filling(move, color)) continue;
#if USE_PRIOR
#if LOW_PRIOR_SELF_ATARI
    uct_init_arc(p, n, move,
	  gg_is_self_atari(move, color) ? Q_SELF_ATARI : 0.5, M_PRIOR);
#elif PRIOR_LIBERTIES
    uct_init_arc(p, n, move, gg_count_libs(move) == 1 ? 0.1 : 0.5, M_PRIOR);
#else
    uct_init_arc(p, n, move, 0.5, M_PRIOR);
#endif
#else
    uct_init_arc(p, n, move, 0, 0);
#endif
    n += 1;
  }
  if(n <= ADD_PASS_THRESHOLD) {
if(0) { gprintf("!%s n < " NAME2STR(ADD_PASS_THRESHOLD)
	      ": p=%p sp=%p empty_list:", __func__, p, sp);
	GG_FOR_ALL_EMPTY_POINT(move) gprintf(" %1m", move); gprintf("\n"); }
#if USE_PRIOR
    uct_init_arc(p, n, GG_PASS_MOVE, Q_PASS, M_PRIOR);
#else
    uct_init_arc(p, n, GG_PASS_MOVE, 0, 0);
#endif
    n += 1;
  }
  p->n_arc = n;
/* Update Q+rave of each move in the node. */
  for(int arc = 0; arc < n; arc += 1) {
#if USE_RAVE
    p->m_rave += p->arc[arc].m_rave;
#endif
    p->visits += p->arc[arc].visits;
  }
  for(int arc = 0; arc < n; arc += 1) {
#if USE_RAVE
    float ucb;
    if(p->arc[arc].visits > 0) {
      float beta;
      ucb = uct_ucb(p, arc);
      if(p->arc[arc].m_rave > 0) {
#if USE_NEW_RAVE	/* TODO */
	const float q_r = uct_ucb_rave(p, arc);
	const float q_u = p->arc[arc].value;
	const float m = p->arc[arc].m_rave;
	const float n = p->arc[arc].visits;
	const float b_r_sqr = SQR(RAVE_BIAS);
	const float q_ur = ;
	const float
	  beta = m / (m + n + m * n * b_r_sqr / (q_ur * (1 - q_ur)));
	const float
	  ucb_r = q_r + C_RAVE * sqrtf(logf(p->m_rave) / m);
	const float
	  ucb_u = q_u + C_UCT * sqrtf(logf(p->visits) / n);
	ucb = beta * ucb_r + (1 - beta) * ucb_u;
#else
	beta = sqrtf(gg_k_rave / (3 * p->visits + gg_k_rave));
	ucb = beta * uct_ucb_rave(p, arc) + (1 - beta) * ucb;
#endif
      }
    } else {
      ucb = gg_fpu_initial;
    }
    p->arc[arc].ucb = ucb;
#else
    p->arc[arc].ucb = gg_fpu_initial;
#endif
  }
if(0) gprintf("<-- %s node=%d n_arc=%d\n", __func__, node, n);
  return &gg_uct_node[node];
}

ggSArc
uct_select_optarc(ggUctNode *p) {
  ggSArc optarc, arc;
  float ucb, ucbopt;

if(0) gprintf("--> %s(%p)\n", __func__, p);
  assert(p->n_arc > 0);
#if USE_ARC_CACHE
  if((optarc = p->optarc) >= 0)	goto exit;
#endif
  /* optarc = argmax(ucb) */
  optarc = 0;
  if(p->n_arc > 1) {
    optarc = -1;
    ucbopt = -1;
    for(arc = 0; arc < p->n_arc; arc += 1) {
      ucb = p->arc[arc].ucb;
      if(ucb > ucbopt) {
	optarc = arc;
	ucbopt = ucb;
      }
    }
  }
#if USE_ARC_CACHE
  p->optarc = optarc;
exit:
#endif
if(0) gprintf("<-- %s arc=%d\n", __func__, optarc);
  return optarc;
}

void
uct_update_tree(ggUctNode *p, ggArc optarc, float value) {
if(0) gprintf("--> %s(%p,%d,%.1f)\n", __func__, p, optarc, value);
  assert(p != NULL);
  do {
    assert(&gg_uct_node[0] <= p && p < &gg_uct_node[uct_n_node]);
#if 1
    assert(optarc < p->n_arc);
#else
    if(!(optarc < p->n_arc)) {
      GG_FATAL("%s: p=%p optarc=%d n_arc=%d\n", __func__, p, optarc, p->n_arc);
    }
#endif
#if USE_ARC_CACHE
    p->optarc = -1;
#endif
    p->arc[optarc].visits += 1;
    p->visits += 1;
    p->arc[optarc].value +=
	    (value - p->arc[optarc].value) / p->arc[optarc].visits;
#if USE_UCB1_TUNED
    p->arc[optarc].value2 +=
      (SQR(value) - p->arc[optarc].value2) / p->arc[optarc].visits;
#endif
#if PROPER_UCT
/* Compute ucb value of all moves in this node
 * This is the proper implementation. */
    for(ggMove arc = 0; arc < p->n_arc; arc += 1) {
      if(p->arc[arc].visits > 0) {
	p->arc[arc].ucb = uct_ucb(p, arc);
      }
    }
#else
    p->arc[optarc].ucb = uct_ucb(p, optarc);
#endif
if(0) gprintf("UCT  uct[%6d].%1m=%+6.3f v=%.0f\n",
	UCT_NTH(p), p->arc[optarc].move, p->arc[optarc].ucb, value);
    value = 1 - value;
    optarc = p->p_arc;
    p = p->p_node;
  } while(p != NULL);
if(0) gprintf("<-- %s\n", __func__);
  return;
}

#if USE_RAVE

static inline float
rave_rt(float v, ggColor c1, ggColor c2) {
  return c1 == c2 ? v : 1 - v;
}

/* Update the values of the nodes according to RAVE algorithm.
 */
void
rave_update_tree(ggUctNode *leaf, ggArc optarc, float value,
		    ggColor color, const mcRave *rave) {
  ggUctNode *p;
  int i;
  ggArc arc;
  ggMove move;
  float v;
  const int n_rave = rave->n_rave;
  ggColor c_move[GG_BOARDSIZE];

if(0) gprintf("--> %s(%p,%d,%.1f,%C,%p)\n", __func__,
			  leaf, optarc, value, color, rave);
if(0) {
  gprintf("%s:", __func__);
  for(i = 0; i < n_rave; i += 1) {
    gprintf(" rave[%d]=%1m(%C)", i, rave->move[i], rave->color[i]);
  }
  gputc('\n');
}

  /* Setup c_move[].
   */
  for(move = 0; move < GG_BOARDSIZE; move += 1) {
    c_move[move] = GG_EMPTY;
  }
  for(i = n_rave; --i >= 0; ) {
    ggColor c;
    move = rave->move[i];
    assert(move < GG_BOARDSIZE);
    c = rave->color[i];
    assert(c == GG_BLACK || c == GG_WHITE);
    c_move[move] = c;
  }

  assert(leaf != NULL);
  for(p = leaf, arc = optarc; p != NULL; arc = p->p_arc, p = p->p_node) {
    c_move[p->arc[arc].move] = p->color;
  }

  for(p = leaf; p != NULL; p = p->p_node) {
    /* Accumulate score. */
    v = rave_rt(value, color, p->color);
    for(arc = 0; arc < p->n_arc; arc += 1) {
      if(c_move[p->arc[arc].move] != p->color)	continue;
      p->m_rave += 1;
      p->arc[arc].m_rave += 1;
      p->arc[arc].q_rave += (v - p->arc[arc].q_rave) / p->arc[arc].m_rave;
#if USE_UCB1_TUNED
      p->arc[arc].q_rave_sqr +=
	    (SQR(v) - p->arc[arc].q_rave_sqr) / p->arc[arc].m_rave;
#endif
    }

    /* Update Q+rave of moves in the node. */
    for(arc = 0; arc < p->n_arc; arc += 1) {
      float beta, ucb_rave;
      if(p->arc[arc].m_rave == 0)	continue;
      beta = sqrtf(gg_k_rave / (3 * p->visits + gg_k_rave));
      ucb_rave = uct_ucb_rave(p, arc);
      p->arc[arc].ucb = beta * ucb_rave + (1 - beta) * p->arc[arc].ucb;
if(0) gprintf("Rave uct[%d].%1m=(%+4.1f/%+4.1f) v=%.0f b=%.0f%%.\n",
	  UCT_NTH(p), move, p->arc[arc].ucb, ucb_rave, 
	  rave_rt(value, color, p->color), beta * 100);
    }
  }
if(0) gprintf("<-- %s\n", __func__);
  return;
}

#endif
